// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"



// boundaries of consideration:
//
//  * physical memory boundary: 1024k (1Megabyte) (increments of...)
//  * A20 boundary:             1024k (1Megabyte)
//  * page boundary:            4k
//  * ROM boundary:             2k (dont care since we are only reading)
//  * segment boundary:         any




// based on the prefetch queue size, define values needed
// to mask out the blocks of memory aligned on boundaries
// of the prefetch queue size.

#if BX_PREFETCH_QSIZE == 8
#  define BX_PREFETCH_BLOCK_MASK      0xfffffff8
#  define BX_PREFETCH_OFFSET_MASK     0x00000007
#elif BX_PREFETCH_QSIZE == 16
#  define BX_PREFETCH_BLOCK_MASK      0xfffffff0
#  define BX_PREFETCH_OFFSET_MASK     0x0000000f
#elif BX_PREFETCH_QSIZE == 32
#  define BX_PREFETCH_BLOCK_MASK      0xffffffe0
#  define BX_PREFETCH_OFFSET_MASK     0x0000001f
#else
#  error "BX_PREFETCH_QSIZE is not supported a supported value"
#endif





  void
bx_cpu_c::prefetch(void)
{
  // cs:eIP
  // prefetch QSIZE byte quantity aligned on corresponding boundary
  Bit32u linear_instr_addr;
  Bit32u phy_block64_addr;

  linear_instr_addr = BX_CPU_THIS_PTR cs.cache.u.segment.base + BX_CPU_THIS_PTR eip;
  BX_CPU_THIS_PTR linear_block_addr = linear_instr_addr & BX_PREFETCH_BLOCK_MASK;
  BX_CPU_THIS_PTR fetch_index = linear_instr_addr & BX_PREFETCH_OFFSET_MASK;
// ??? put eip check here.  take out below, after decode step is unified.

  if (BX_CPU_THIS_PTR cr0.pg) {
    // aligned block guarenteed to be all in one page, same A20 address
    phy_block64_addr = translate_linear(BX_CPU_THIS_PTR linear_block_addr, CPL==3, BX_READ);
    phy_block64_addr = A20ADDR(phy_block64_addr);
    }
  else {
    phy_block64_addr = A20ADDR(BX_CPU_THIS_PTR linear_block_addr);
    }

  if ( phy_block64_addr >= bx_mem.len )
    return; // ??? maybe panic here

  // dont really care about endian-ness here - just move 4bytes at a time, in same order
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[0])  = *( (Bit32u *) &bx_mem.vector[phy_block64_addr]);
  BX_MEMORY_HOOK (phy_block64_addr, 4, BX_READ, NULL);
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[4])  = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 4]);
  BX_MEMORY_HOOK (phy_block64_addr+4, 4, BX_READ, NULL);
#if BX_PREFETCH_QSIZE > 8
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[8])  = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 8]);
  BX_MEMORY_HOOK (phy_block64_addr+8, 4, BX_READ, NULL);
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[12]) = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 12]);
  BX_MEMORY_HOOK (phy_block64_addr+12, 4, BX_READ, NULL);
#if BX_PREFETCH_QSIZE > 16
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[16]) = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 16]);
  BX_MEMORY_HOOK (phy_block64_addr+16, 4, BX_READ, NULL);
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[20]) = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 20]);
  BX_MEMORY_HOOK (phy_block64_addr+20, 4, BX_READ, NULL);
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[24]) = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 24]);
  BX_MEMORY_HOOK (phy_block64_addr+24, 4, BX_READ, NULL);
  *( (Bit32u *) &BX_CPU_THIS_PTR prefetch_queue[28]) = *( (Bit32u *) &bx_mem.vector[phy_block64_addr + 28]);
  BX_MEMORY_HOOK (phy_block64_addr+28, 4, BX_READ, NULL);
#endif
#endif
}

  // If control has transfered locally, it is possible the prefetch Q is
  // still valid.  This would happen for repeat instructions, and small
  // branches.
  void
bx_cpu_c::revalidate_prefetch_q(void)
{
  Bit32u linear_instr_addr;
  Bit32u temp_linear_block_addr;

  linear_instr_addr = BX_CPU_THIS_PTR cs.cache.u.segment.base + BX_CPU_THIS_PTR eip;
  temp_linear_block_addr = linear_instr_addr & BX_PREFETCH_BLOCK_MASK;
  if (temp_linear_block_addr == BX_CPU_THIS_PTR linear_block_addr) {
    BX_CPU_THIS_PTR fetch_index = linear_instr_addr & BX_PREFETCH_OFFSET_MASK;
    }
  else
    BX_CPU_THIS_PTR fetch_index = BX_INVALID_FETCH_INDEX;
}

  INLINE void
bx_cpu_c::invalidate_prefetch_q(void)
{
  BX_CPU_THIS_PTR fetch_index = BX_INVALID_FETCH_INDEX;
}


  INLINE Bit8u
bx_cpu_c::fetch_next_byte(void)
{
  if (BX_CPU_THIS_PTR eip > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
    bx_panic("fetch_next_byte: EIP > CS.limit\n");
    }

  if (BX_CPU_THIS_PTR fetch_index < BX_PREFETCH_QSIZE) {
    BX_CPU_THIS_PTR eip += 1;
    return( BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++] );
    }
  prefetch();
  BX_CPU_THIS_PTR eip += 1;
  return( BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++] );
}


  INLINE Bit8u
bx_cpu_c::peek_next_byte(void)
{
  if (BX_CPU_THIS_PTR eip > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
    bx_panic("peek_next_byte: EIP > CS.limit\n");
    }

  if (BX_CPU_THIS_PTR fetch_index < BX_PREFETCH_QSIZE) {
    return( BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index] );
    }
  prefetch();
  return( BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index] );
}


  INLINE Bit16u
bx_cpu_c::fetch_next_word(void)
{
  Bit8u b_low, b_high;

  if (BX_CPU_THIS_PTR eip >= BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
    bx_panic("fetch_next_word: EIP+1 > CS.limit\n");
    }

  if (BX_CPU_THIS_PTR fetch_index < (BX_PREFETCH_QSIZE-1)) {
    b_low  = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    b_high = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    BX_CPU_THIS_PTR eip += 2;
    return( (b_high<<8) | b_low );
    }
  else if (BX_CPU_THIS_PTR fetch_index == (BX_PREFETCH_QSIZE-1)) {
    b_low  = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++]; // ??? remove increment here
    BX_CPU_THIS_PTR eip += 1;
    prefetch();
    BX_CPU_THIS_PTR eip += 1;
    b_high = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    return( (b_high<<8) | b_low );
    }
  else {
    prefetch();
    b_low  = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    b_high = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    BX_CPU_THIS_PTR eip += 2;
    return( (b_high<<8) | b_low );
    }
}

  INLINE Bit32u
bx_cpu_c::fetch_next_dword(void)
{
  Bit8u b[4];
  unsigned index;

  if ((BX_CPU_THIS_PTR eip+3) > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
    bx_panic("fetch_next_word: EIP+1 > CS.limit\n");
    }

// optimize this
  if (BX_CPU_THIS_PTR fetch_index < (BX_PREFETCH_QSIZE-3)) {
    b[0] = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    b[1] = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    b[2] = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    b[3] = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
    BX_CPU_THIS_PTR eip += 4;
    return( (b[3]<<24) | (b[2]<<16) | (b[1]<<8) | b[0] );
    }
  else {
    // code spans multiple QSIZE byte blocks
    for (index=0; BX_CPU_THIS_PTR fetch_index<BX_PREFETCH_QSIZE; index++) {
      b[index] = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
      BX_CPU_THIS_PTR eip += 1;
      }
    prefetch();
    for (;index<4; index++) {
      b[index] = BX_CPU_THIS_PTR prefetch_queue[BX_CPU_THIS_PTR fetch_index++];
      BX_CPU_THIS_PTR eip += 1;
      }
    return( (b[3]<<24) | (b[2]<<16) | (b[1]<<8) | b[0] );
    }
}
