#include <stdio.h>
#include <string.h>
#include <signal.h>

#include "bochs.h"


// Change these to point to the C function which
// initializes each simulation environment
#define BX_SIM1_INIT_FUNCTION    bx_bochs_init
#define BX_SIM2_INIT_FUNCTION    sim2_init

#define BX_SIM1_EXIT(code)   bx_panic("debugger exiting.\n")
#define BX_SIM2_EXIT(code)   sim2_exit(code)

// This section for declarations of functions from other simulation
// environments, which are declared in macros above.
// BEGIN ==========================================================

// Put C++ declarations here.

// Put C declarations here.
extern "C" {
  void sim2_init(int argc, char *argv[]);
  void sim2_exit(int code);
  };
// END ============================================================

#define BX_INFILE_DEPTH  10

#define INCLUDE_CMD       "source"
#define INCLUDE_CMD_LEN   strlen(INCLUDE_CMD)


void bx_dbg_usage(char *);

static char bx_debug_rc_fname[BX_MAX_PATH];

static char tmp_buf[512];
static char *tmp_buf_ptr;

static char argv0[BX_MAX_PATH];


typedef struct {
  FILE    *fp;
  char     fname[BX_MAX_PATH];
  unsigned lineno;
  } bx_infile_stack_entry_t;

bx_infile_stack_entry_t bx_infile_stack[BX_INFILE_DEPTH];
int                     bx_infile_stack_index = 0;

static void bx_get_command(void);
static int  bx_nest_infile(char *path);
static void bx_unnest_infile(void);

static void bx_debug_ctrlc_handler(int signum);
volatile Boolean bx_dbg_interrupt_requested = 0;


static void bx_dbg_cpu_loop(void);

static bx_bpoint_state_t bpoint_state[2];
bx_bpoint_requests_t bx_dbg_bpoint_requests;
static void bx_dbg_init_breakpoints(void);
static Bit8u bx_disasm_ibuf[32];
#if BX_DISASM
static char bx_disasm_tbuf[512];
#endif

// for debugging the disassembly stuff only
#define BX_DISASM_CHECK 0

#if BX_DISASM_CHECK
  void bx_dbg_disasm_check(void);
#endif


  int
bx_dbg_main(int argc, char *argv[])
{
  int i;
  unsigned sim1_arg_index=0, sim2_arg_index=0;
  int      sim1_argc=0, sim2_argc=0;
  int reti;

  strncpy(argv0, argv[0], BX_MAX_PATH);
  argv0[BX_MAX_PATH - 1] = 0;

  bx_debug_rc_fname[0] = '\0';

  for (i=1; i<argc; i++) {

    if (strcmp(argv[i], "-sim1") == 0) {
      sim1_arg_index = i;
      sim1_argc = 1;
      for (i=i+1; i<argc; i++) {
        if (strcmp(argv[i], "-sim2") == 0) {
          i--; // backup i
          break;
          }
        sim1_argc++;
        }
      }

    else if (strcmp(argv[i], "-sim2") == 0) {
      sim2_arg_index = i;
      sim2_argc = 1;
      for (i=i+1; i<argc; i++) {
        if (strcmp(argv[i], "-sim1") == 0) {
          i--; // backup i
          break;
          }
        sim2_argc++;
        }
      }

    else if (strcmp(argv[i], "-rc") == 0) {
      if ( (i+1) >= argc ) {
        fprintf(stderr, "%s: -rc option used, but no path specified.\n",
          argv[0]);
        bx_dbg_usage(argv[0]);
        exit(1);
        }
      strncpy(bx_debug_rc_fname, argv[i+1], BX_MAX_PATH-1);
      i += 1; // increment one extra for the path
      }

    else {
      fprintf(stderr, "%s: %s option not understood.\n",
        argv[0], argv[i]);
      bx_dbg_usage(argv[0]);
      exit(1);
      }
    }

  bx_infile_stack_index = 0;
  bx_infile_stack[0].fp = stdin;
  strncpy(bx_infile_stack[0].fname, argv[0], BX_MAX_PATH);
  bx_infile_stack[0].fname[BX_MAX_PATH-1] = 0;
  bx_infile_stack[0].lineno = 0;


  if (bx_debug_rc_fname[0] == '\0') {
    fprintf(stderr, "%s: Warning: no rc file specified.\n", argv[0]);
    }
  else {
    fprintf(stderr, "%s: using rc file '%s'.\n", argv[0], bx_debug_rc_fname);
    // if there's an error, the user will know about it before proceeding
    (void) bx_nest_infile(bx_debug_rc_fname);
    }

  bx_dbg_init_breakpoints();

  for (unsigned j=0; j<sizeof(bx_disasm_ibuf); j++)
    bx_disasm_ibuf[j] = 0;

  BX_SIM1_INIT_FUNCTION (sim1_argc, &argv[sim1_arg_index]);
  BX_SIM2_INIT_FUNCTION (sim2_argc, &argv[sim2_arg_index]);

  // setup Ctrl-C handler
  signal(SIGINT, bx_debug_ctrlc_handler);

#if BX_DISASM_CHECK
  bx_dbg_disasm_check();
#endif

  while ( 1 ) {
    bx_get_command();
    if ( (*tmp_buf_ptr == '\n') || (*tmp_buf_ptr == 0) ) {
      fprintf(stderr, "\n");
      }
    else if ( (strncmp(tmp_buf_ptr, INCLUDE_CMD, INCLUDE_CMD_LEN) == 0) &&
              (tmp_buf_ptr[INCLUDE_CMD_LEN] == ' ' ||
               tmp_buf_ptr[INCLUDE_CMD_LEN] == '\t') ) {
      char *ptr;
      int len;

      ptr = tmp_buf_ptr + INCLUDE_CMD_LEN+1;
      while ( *ptr==' ' || *ptr=='\t' )
        ptr++;

      len = strlen(ptr);
      if (len == 0) {
        fprintf(stderr, "%s: no filename given to 'source' command.\n", argv0);
        if (bx_infile_stack_index > 0) {
          fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
          bx_dbg_exit(1);
          }
        continue;
        }
      ptr[len-1] = 0; // get rid of newline
      fprintf(stderr, "\n");
      reti = bx_nest_infile(ptr);
      if ((reti==0) && (bx_infile_stack_index > 0)) {
        fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
        bx_dbg_exit(1);
        }
      }
    else {
      // Give a chance to the command line extensions, to
      // consume the command.  If they return 0, then
      // we need to process the command.  A return of 1
      // means, the extensions have handled the command
      if ( bx_dbg_extensions(tmp_buf_ptr)==0 ) {
        // process command here
        //fprintf(stderr, "read line: '%s'\n", tmp_buf_ptr);
        bx_add_lex_input(tmp_buf_ptr);
        yyparse();
        }
      }
    }

  bx_dbg_exit(0);
}


  void
bx_get_command(void)
{
  char *charptr_ret;

  bx_infile_stack[bx_infile_stack_index].lineno++;

  fprintf(stderr, "%s:%d> ", bx_infile_stack[bx_infile_stack_index].fname,
    bx_infile_stack[bx_infile_stack_index].lineno);

  charptr_ret = fgets(tmp_buf, 512,
    bx_infile_stack[bx_infile_stack_index].fp);
  if (charptr_ret == NULL) {
    // see if error was due to EOF condition
    if (feof(bx_infile_stack[bx_infile_stack_index].fp)) {
      if (bx_infile_stack_index > 0) {
        // nested level of include files, pop back to previous one
        fprintf(stderr, "\n");
        bx_unnest_infile();
        }
      else {
        // not nested, sitting at stdin prompt, user wants out
        fprintf(stderr, "%s: BYE.\n", argv0);
        bx_dbg_exit(0);
        }

      // call recursively
      bx_get_command();
      return;
      }

    // error was not EOF, see if it was from a Ctrl-C
    if (bx_dbg_interrupt_requested) {
      tmp_buf[0] = '\n';
      tmp_buf[1] = 0;
      tmp_buf_ptr = &tmp_buf[0];
      bx_dbg_interrupt_requested = 0;
      return;
      }

    fprintf(stderr, "fgets() returned ERROR.\n");
fprintf(stderr, "intr request was %u\n", bx_dbg_interrupt_requested);
    bx_dbg_exit(1);
    }
  tmp_buf_ptr = &tmp_buf[0];

  // look for first non-whitespace character
  while ( ((*tmp_buf_ptr == ' ') || (*tmp_buf_ptr == '\t')) &&
          (*tmp_buf_ptr != '\n') && (*tmp_buf_ptr != 0) ) {
    tmp_buf_ptr++;
    }

  return;
}


  void
bx_dbg_usage(char *progname)
{
  fprintf(stderr, "usage: %s [-rc path] [-sim1 ... ] [-sim2 ... ]\n", progname);
}

  void
bx_dbg_continue(void)
{
  // continue executing, until a breakpoint or request occurs

  bx_dbg_bpoint_requests.instr_count_max = 0;  // unlimited
  bx_dbg_cpu_loop();
}

  void
bx_dbg_single_step(void)
{
  // next instruction, stepping into function calls

  bx_dbg_bpoint_requests.instr_count_max = 1;  // one instruction
  bx_dbg_cpu_loop();
}


  void
bx_dbg_cpu_loop(void)
{
  // zero out breakpoint request structure
  memset(&bpoint_state[0], 0, sizeof(bpoint_state[0]));
  memset(&bpoint_state[1], 0, sizeof(bpoint_state[1]));

  // execute simulator until a requested condition occurs
  fprintf(stderr, "cpu_loop called.\n");
  bx_dbg_interrupt_requested = 0;
  bx_cpu.cpu_loop();
  bx_dbg_interrupt_requested = 0;

  if (bpoint_state[0].before.iaddr.phy_addr_valid) {
#if BX_DISASM
    unsigned ilen;

    bx_dbg_fetch_mem(bpoint_state[0].before.iaddr.phy_addr, 16, bx_disasm_ibuf);
    ilen = bx_disassemble.disasm(bpoint_state[0].before.is_32bit_code,
                                 bx_disasm_ibuf, bx_disasm_tbuf);
 
    if (bpoint_state[0].before.is_32bit_code) {
      fprintf(stderr, "%04x:%08x: ",
              (unsigned) bpoint_state[0].before.iaddr.cs,
              (unsigned) bpoint_state[0].before.iaddr.eip);
      }
    else {
      fprintf(stderr, "%04x:%04x: ",
              (unsigned) bpoint_state[0].before.iaddr.cs,
              (unsigned) bpoint_state[0].before.iaddr.eip);
      }
    for (unsigned j=0; j<ilen; j++)
      fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
    fprintf(stderr, ": %s\n", bx_disasm_tbuf);
#endif  // #if BX_DISASM
    }
  else {
    fprintf(stderr, "??? (physical address not available)\n");
    }

#if 0
  fprintf(stderr, "cpu_loop returned (%u).\n",
    (unsigned) bpoint_state[0].instr_count);
  fprintf(stderr, "  cs:eip = %04x:%08x, %04x:%08x\n",
     (unsigned) bpoint_state[0].before.iaddr.cs,
     (unsigned) bpoint_state[0].before.iaddr.eip,
     (unsigned) bpoint_state[0].after.iaddr.cs,
     (unsigned) bpoint_state[0].after.iaddr.eip);

  fprintf(stderr, "  is_instr_count_bpoint = %u\n",
    bpoint_state[0].is_instr_count_bpoint);
  fprintf(stderr, "  is_vir_iaddr_bpoint = %u\n",
    bpoint_state[0].is_vir_iaddr_bpoint);
  fprintf(stderr, "  is_lin_iaddr_bpoint = %u\n",
    bpoint_state[0].is_lin_iaddr_bpoint);
  fprintf(stderr, "  is_phy_iaddr_bpoint = %u\n",
    bpoint_state[0].is_phy_iaddr_bpoint);
  fprintf(stderr, "  is_io_bpoint = %u\n",
    bpoint_state[0].is_io_bpoint);
  fprintf(stderr, "  is_soft_int_bpoint = %u\n",
    bpoint_state[0].is_soft_int_bpoint);
  fprintf(stderr, "  is_hard_int_bpoint = %u\n",
    bpoint_state[0].is_hard_int_bpoint);
  fprintf(stderr, "  is_memory_bpoint = %u\n",
    bpoint_state[0].is_memory_bpoint);
  fprintf(stderr, "  is_opcode_bpoint = %u\n",
    bpoint_state[0].is_opcode_bpoint);
  fprintf(stderr, "  is_exception_bpoint = %u\n",
    bpoint_state[0].is_exception_bpoint);
  fprintf(stderr, "  is_user_interrupt = %u\n",
    bpoint_state[0].is_user_interrupt);
#endif
}


  void
bx_dbg_set(char *p1, char *p2, char *p3)
{
  fprintf(stderr, "%s %s %s: command 'set' not implemented yet.\n",
    p1, p2, p3);
}

  void
bx_dbg_breakpoint(char *p1, char *p2)
{
  fprintf(stderr, "%s %s: command 'b' not implemented yet.\n",
    p1, p2);
}

  void
bx_dbg_del_breakpoint(unsigned handle)
{
  if (handle >= bx_dbg_bpoint_requests.num_iaddr_bpoints) {
    fprintf(stderr, "No breakpoint number %u.\n", handle);
    return;
    }

  bx_dbg_bpoint_requests.bpoint_type[handle] = BX_BPOINT_TYPE_NONE;
}

  int
bx_nest_infile(char *path)
{
  FILE *tmp_fp;
 
  tmp_fp = fopen(path, "r");
  if (!tmp_fp) {
    fprintf(stderr, "%s: can not open file '%s' for reading.\n",
            argv0, path);
    return(0);
    }
 
  if ( (bx_infile_stack_index+1) >= BX_INFILE_DEPTH ) {
    fprintf(stderr, "%s: source files nested too deeply\n", argv0);
    return(0);
    }
 
  bx_infile_stack_index++;
  bx_infile_stack[bx_infile_stack_index].fp = tmp_fp;
  strncpy(bx_infile_stack[bx_infile_stack_index].fname, path, BX_MAX_PATH);
  bx_infile_stack[bx_infile_stack_index].fname[BX_MAX_PATH-1] = 0;
  bx_infile_stack[bx_infile_stack_index].lineno = 0;
  return(1);
}

  void
bx_unnest_infile(void)
{
  if (bx_infile_stack_index <= 0) {
    fprintf(stderr, "%s: ERROR: unnest_infile(): nesting level = 0.\n",
      argv0);
    bx_dbg_exit(1);
    }

  fclose(bx_infile_stack[bx_infile_stack_index].fp);
  bx_infile_stack_index--;
}

  int
yywrap(void)
{
  fprintf(stderr, "%s: ERROR: yywrap() called.\n", argv0);
  bx_dbg_exit(1);
  return(0); // keep compiler quiet
}


  void
yyerror(char *s)
{
  fprintf(stderr, "%s:%d: %s at '%s'\n",
    bx_infile_stack[bx_infile_stack_index].fname,
    bx_infile_stack[bx_infile_stack_index].lineno,
    s, yytext);

  if (bx_infile_stack_index > 0) {
    fprintf(stderr, "%s: ERROR in source file causes exit.\n", argv0);
    bx_dbg_exit(1);
    }
}

  void
bx_debug_ctrlc_handler(int signum)
{
  UNUSED(signum);

  signal(SIGINT, bx_debug_ctrlc_handler);
  bx_dbg_interrupt_requested = 1;
}

  void
bx_dbg_exit(int code)
{
fprintf(stderr, "before sim1_exit\n");
 bx_panic("debugger exiting.\n");
  //BX_SIM1_EXIT(code);
fprintf(stderr, "before sim2_exit\n");
  BX_SIM2_EXIT(code);
fprintf(stderr, "before exit\n");

  exit(code);
}

  void
bx_dbg_vbreakpoint(Boolean specific, Bit32u cs, Bit32u eip)
{
  unsigned int i = bx_dbg_bpoint_requests.num_iaddr_bpoints;

  if (cs > 0xffff) {
    fprintf(stderr, "%s: virtual breakpoint: CS greater than 0xFFFF.\n", argv0);
    return;
    }

  if (specific) {
    if ( (i+1) <= BX_MAX_IADDR_BPOINTS) {
      bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_VIRTUAL;
      bx_dbg_bpoint_requests.iaddr[i].vir.cs  = cs;
      bx_dbg_bpoint_requests.iaddr[i].vir.eip = eip;
      bx_dbg_bpoint_requests.attr[i].disposition = bx_disp_keep;
      bx_dbg_bpoint_requests.attr[i].enabled = 1;
      bx_dbg_bpoint_requests.attr[i].ignorecount = 0;
      bx_dbg_bpoint_requests.num_iaddr_bpoints++;
      }
    else {
      fprintf(stderr, "%s: virtual breakpoint: no bpoint slots left.\n", argv0);
      }
    }
  else
    fprintf(stderr, "VBREAK not implemented (not-specific)\n");
}


  void
bx_dbg_lbreakpoint(Boolean specific, Bit32u laddress)
{
  unsigned int i = bx_dbg_bpoint_requests.num_iaddr_bpoints;

  if (specific) {
    if ( (i+1) <= BX_MAX_IADDR_BPOINTS) {
      bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_LINEAR;
      bx_dbg_bpoint_requests.iaddr[i].lin.addr = laddress;
      bx_dbg_bpoint_requests.attr[i].disposition = bx_disp_keep;
      bx_dbg_bpoint_requests.attr[i].enabled = 1;
      bx_dbg_bpoint_requests.attr[i].ignorecount = 0;
      bx_dbg_bpoint_requests.num_iaddr_bpoints++;
      }
    else {
      fprintf(stderr, "%s: linear breakpoint: no bpoint slots left.\n", argv0);
      }
    }
  else
    fprintf(stderr, "LBREAK not implemented (not-specific)\n");
}

  void
bx_dbg_pbreakpoint(Boolean specific, Bit32u paddress)
{
  unsigned int i = bx_dbg_bpoint_requests.num_iaddr_bpoints;

  if (specific) {
    if ( (i+1) <= BX_MAX_IADDR_BPOINTS) {
      bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_PHYSICAL;
      bx_dbg_bpoint_requests.iaddr[i].phy.addr = paddress;
      bx_dbg_bpoint_requests.attr[i].disposition = bx_disp_keep;
      bx_dbg_bpoint_requests.attr[i].enabled = 1;
      bx_dbg_bpoint_requests.attr[i].ignorecount = 0;
      bx_dbg_bpoint_requests.num_iaddr_bpoints++;
      }
    else {
      fprintf(stderr, "%s: physical breakpoint: no bpoint slots left.\n", argv0);
      }
    }
  else
    fprintf(stderr, "PBREAK not implemented (not-specific)\n");
}


  void
bx_dbg_init_breakpoints(void)
{
  unsigned i;
  bx_dbg_bpoint_requests.instr_count_max = 0;

  bx_dbg_bpoint_requests.num_iaddr_bpoints = 0;
  for (i=0; i<BX_MAX_IADDR_BPOINTS; i++) {
    bx_dbg_bpoint_requests.bpoint_type[i] = BX_BPOINT_TYPE_NONE;
    }
}


  Boolean
bx_dbg_cpu_icount(Bit32u cs, Bit32u eip,
                  Bit32u laddr, Bit32u paddr, Boolean paddr_valid,
                  Boolean intr_pending, Boolean is_32bit_code)
{
  // function returns
  //   0: CPU should continue executing
  //   1: CPU should return immediately to debug code


  // if any break condition has occurred DURING execution of
  // last instruction, store state info about next instruction,
  // and tell CPU to return to the debug environment
  if (bpoint_state[0].event_occurred) {
    bpoint_state[0].after.intr_pending = intr_pending;
    bpoint_state[0].after.iaddr.cs = cs;
    bpoint_state[0].after.iaddr.eip = eip;
    bpoint_state[0].after.iaddr.linear_addr = laddr;
    bpoint_state[0].after.iaddr.phy_addr = paddr;
    bpoint_state[0].after.iaddr.phy_addr_valid = paddr_valid;
    bpoint_state[0].after.is_32bit_code = is_32bit_code;
    return(1);
    }

  // store state info about current instruction to be executed next
  bpoint_state[0].before.intr_pending = intr_pending;
  bpoint_state[0].before.iaddr.cs = cs;
  bpoint_state[0].before.iaddr.eip = eip;
  bpoint_state[0].before.iaddr.linear_addr = laddr;
  bpoint_state[0].before.iaddr.phy_addr = paddr;
  bpoint_state[0].before.iaddr.phy_addr_valid = paddr_valid;
  bpoint_state[0].before.is_32bit_code = is_32bit_code;

  //
  // Now look for break conditions BEFORE execution of next instruction
  //

  // look for Ctrl-C
  if (bx_dbg_interrupt_requested) {
    bpoint_state[0].is_user_interrupt = 1;
    bpoint_state[0].event_occurred = 1;
    }

  for (unsigned i=0; i<bx_dbg_bpoint_requests.num_iaddr_bpoints; i++) {

    if ( bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_VIRTUAL ) {
      // look for breakpoints at this instruction's virtual address
      if ( (bx_dbg_bpoint_requests.iaddr[i].vir.cs  == cs) &&
           (bx_dbg_bpoint_requests.iaddr[i].vir.eip == eip) ) {
        if (bpoint_state[0].instr_count==0) continue;
        bpoint_state[0].event_occurred = 1;
        bpoint_state[0].is_vir_iaddr_bpoint = 1;
        bpoint_state[0].vir_iaddr_handle = i;
        }
      }

    else if ( bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_LINEAR ) {
      // look for breakpoints at this instruction's linear address
      if ( bx_dbg_bpoint_requests.iaddr[i].lin.addr  == laddr ) {
        if (bpoint_state[0].instr_count==0) continue;
        bpoint_state[0].event_occurred = 1;
        bpoint_state[0].is_lin_iaddr_bpoint = 1;
        bpoint_state[0].lin_iaddr_handle = i;
        }
      }

    else if ( bx_dbg_bpoint_requests.bpoint_type[i] == BX_BPOINT_TYPE_PHYSICAL ) {
      // look for breakpoints at this instruction's physical address
      if ( bx_dbg_bpoint_requests.iaddr[i].phy.addr  == paddr ) {
        if (bpoint_state[0].instr_count==0) continue;
        bpoint_state[0].event_occurred = 1;
        bpoint_state[0].is_phy_iaddr_bpoint = 1;
        bpoint_state[0].phy_iaddr_handle = i;
        }
      }
    }


  // Look for a step count
  if (bx_dbg_bpoint_requests.instr_count_max &&
      ((bpoint_state[0].instr_count+1) > bx_dbg_bpoint_requests.instr_count_max)) {
    bpoint_state[0].is_instr_count_bpoint = 1;
    bpoint_state[0].event_occurred = 1;
    }

  if (bpoint_state[0].event_occurred) return(1);

  bpoint_state[0].instr_count++;
  return(0);
}

  void
bx_dbg_info_bpoints(void)
{
// Num Type           Disp Enb Address    What
// 1   breakpoint     keep y   0x00010664 in main at temp.c:7

  fprintf(stderr, "Num Type           Disp Enb Address\n");
  for (unsigned i=0; i<bx_dbg_bpoint_requests.num_iaddr_bpoints; i++) {
    if (bx_dbg_bpoint_requests.bpoint_type[i] != BX_BPOINT_TYPE_NONE) {
      fprintf(stderr, "%3u ", i);
      switch (bx_dbg_bpoint_requests.bpoint_type[i]) {
        case BX_BPOINT_TYPE_VIRTUAL:  fprintf(stderr, "vbreakpoint    "); break;
        case BX_BPOINT_TYPE_LINEAR:   fprintf(stderr, "lbreakpoint    "); break;
        case BX_BPOINT_TYPE_PHYSICAL: fprintf(stderr, "pbreakpoint    "); break;
        }
      switch (bx_dbg_bpoint_requests.attr[i].disposition) {
        case bx_disp_disable: fprintf(stderr, "dis  "); break;
        case bx_disp_keep:    fprintf(stderr, "keep "); break;
        case bx_disp_delete:  fprintf(stderr, "del  "); break;
        }
      if (bx_dbg_bpoint_requests.attr[i].enabled)
        fprintf(stderr, "y   ");
      else
        fprintf(stderr, "n   ");

      switch (bx_dbg_bpoint_requests.bpoint_type[i]) {
        case BX_BPOINT_TYPE_VIRTUAL:
          fprintf(stderr, "0x%04x:0x%08x\n",
                  bx_dbg_bpoint_requests.iaddr[i].vir.cs,
                  bx_dbg_bpoint_requests.iaddr[i].vir.eip);
          break;
        case BX_BPOINT_TYPE_LINEAR:
          fprintf(stderr, "0x%08x\n",
                  bx_dbg_bpoint_requests.iaddr[i].lin.addr);
          break;
        case BX_BPOINT_TYPE_PHYSICAL:
          fprintf(stderr, "0x%08x\n",
                  bx_dbg_bpoint_requests.iaddr[i].phy.addr);
          break;
        }
      }
    }
}


#if BX_DISASM_CHECK
  void
bx_dbg_disasm_check(void)
{
  unsigned ilen;
  unsigned or16=0, or32=0; // use segment override in 16 & 32bit modes (0 or 1)

  fprintf(stderr, "\nDISASM CHECK: 16bit address mode\n");

  bx_disasm_ibuf[0] = 0x26;      // ES: (if or16==0, not used)
  bx_disasm_ibuf[0+or16] = 0x8a; // mov_gbeb AL, ...
  bx_disasm_ibuf[2+or16] = 0x11; // 8bit address data if needed
  bx_disasm_ibuf[3+or16] = 0x22; // 16bit address data if needed

  for (unsigned mod=0; mod<=3; mod++) {
    for (unsigned rm=0; rm<=7; rm++) {
      bx_disasm_ibuf[1+or16] = (mod << 6) | rm;
      ilen = bx_disassemble.disasm(0, bx_disasm_ibuf, bx_disasm_tbuf);
 
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    }

  fprintf(stderr, "\nDISASM CHECK: 32bit address mode\n");
  fprintf(stderr, "\nOne byte opcodes:\n");

  bx_disasm_ibuf[0] = 0x26;      // ES: (if or32==0, not used)
  bx_disasm_ibuf[0+or32] = 0x67; // ADDRSIZE
  bx_disasm_ibuf[1+or32] = 0x8a; // mov_gbeb AL, ...
  bx_disasm_ibuf[3+or32] = 0x11; // address data bits  0.. 7
  bx_disasm_ibuf[4+or32] = 0x22; // address data bits  8..15
  bx_disasm_ibuf[5+or32] = 0x33; // address data bits 16..23
  bx_disasm_ibuf[6+or32] = 0x44; // address data bits 24..31


  for (unsigned mod=0; mod<=3; mod++) {
    for (unsigned rm=0; rm<=7; rm++) {
      if ( (rm==4) && (mod!=3) ) continue;
      bx_disasm_ibuf[2+or32] = (mod << 6) | rm;
      ilen = bx_disassemble.disasm(0, bx_disasm_ibuf, bx_disasm_tbuf);
 
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    }

  fprintf(stderr, "\nTwo byte opcodes:\n");

  bx_disasm_ibuf[0] = 0x26;      // ES: (if or32==0, not used)
  bx_disasm_ibuf[0+or32] = 0x67; // ADDRSIZE
  bx_disasm_ibuf[1+or32] = 0x8a; // mov_gbeb AL, ...
  bx_disasm_ibuf[4+or32] = 0x11; // address data bits  0.. 7
  bx_disasm_ibuf[5+or32] = 0x22; // address data bits  8..15
  bx_disasm_ibuf[6+or32] = 0x33; // address data bits 16..23
  bx_disasm_ibuf[7+or32] = 0x44; // address data bits 24..31

  for (unsigned mod=0; mod<=2; mod++) {
    unsigned rm, ss, index;

    rm = 4;
    bx_disasm_ibuf[2+or32] = (mod << 6) | rm;
    for (unsigned base=0; base<=7; base++) {
      ss = 3;
      index = 1;
      bx_disasm_ibuf[3+or32] = (ss << 6) | (index<<3) | base;
      ilen = bx_disassemble.disasm(0, bx_disasm_ibuf, bx_disasm_tbuf);
 
      for (unsigned j=0; j<ilen; j++)
        fprintf(stderr, "%02x", (unsigned) bx_disasm_ibuf[j]);
      fprintf(stderr, ": %s\n", bx_disasm_tbuf);
      }
    }
}
#endif
